#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Tests for solicitor.py
"""
# pylint: disable=E0401, C0301, C0116, R0201, W0621
from unittest.mock import patch

import pytest
from ingest_envoy.solicitor import Solicitor
from ingest_envoy.utilities import IngestType

SOLICIT_WORKFLOWS_PATCH = "ingest_envoy.solicitor.Solicitor.solicit_contents"
filename = "test/input_files/test-metadata.json"
image_filename = "test/input_files/test-image-metadata.json"


@pytest.fixture(scope="function")
def solicitor() -> Solicitor:
    return Solicitor(IngestType.CAL, ["http://capability:3457", "http://workflow:3456"], filename=filename)


@pytest.fixture(scope="function")
def image_solicitor() -> Solicitor:
    return Solicitor(IngestType.IMG, ["http://capability:3457", "http://workflow:3456"], filename=image_filename)


expected_metadata = {
    "fileSetIds": "16B-069_sb32814386_1_001.57685.66193635417",
    "workflowName": "std_calibration",
    "systemId": "2",
    "creationTime": "2021-07-06T21:50:48",
    "productLocator": "uid://evla/execblock/48ba4c9d-d7c7-4a8f-9803-1115cd52459b",
    "projectMetadata": {
        "projectCode": "16B-069",
        "title": "JVLA monitoring of an in-progress extreme scattering event",
        "telescope": "EVLA",
        "startTime": "57685.661952546296",
        "observer": "Keith Bannister",
    },
    "destinationDirectory": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmp576dknjp",
}


class TestSolicitor:
    def test_solicit_contents(self, solicitor: Solicitor):
        metadata = solicitor.solicit_contents()
        assert metadata == expected_metadata

    def test_solicit_workflow_directory_name(self, solicitor: Solicitor):
        expected_workflow_dir_name = "tmp576dknjp"

        name = solicitor.solicit_workflow_directory_name()
        assert name == expected_workflow_dir_name

    def test_solicit_processing_creation_time(self, solicitor: Solicitor):
        expected_corrected_time = "2021_07_06T21_50_48"

        time_string = solicitor.solicit_processing_creation_time()
        assert time_string == expected_corrected_time

    def test_solicit_parameters_std_cal(self, solicitor: Solicitor):
        expected_parameters = {
            "sdmId": "16B-069_sb32814386_1_001.57685.66193635417",
            "telescope": "EVLA",
            "workflowName": "std_calibration",
            "spl": "uid://evla/execblock/48ba4c9d-d7c7-4a8f-9803-1115cd52459b",
            "processingStart": "2021_07_06T21_50_48",
            "workflowDir": "tmp576dknjp",
            "initialVersionDir": "tmp24i4o1yl",
            "multiVersion": True,
        }
        with patch("ingest_envoy.solicitor.Solicitor.solicit_initial_directory_name", return_value="tmp24i4o1yl"):
            parameters = solicitor.solicit_parameters()
            assert parameters == expected_parameters

    def test_solicit_parameters_image(self, image_solicitor: Solicitor):
        initial_version_dir = "iamthefirst"
        metadata = {
            "telescope": "EVLA",
            "workflowName": "std_cms_imaging",
            "processingStart": "2021_07_29T14_26_31",
            "workflowDir": "tmprb1se376",
            "initialVersionDir": "iamthefirst",
            "multiVersion": True,
            "calSpl": "uid://evla/calibration/c78ccfd6-fe4e-43c6-a5c5-70e5bcfde16b",
        }
        with patch("ingest_envoy.solicitor.Solicitor.solicit_initial_directory_name", return_value=initial_version_dir):
            parameters = image_solicitor.solicit_parameters()
            assert parameters == metadata
